﻿using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Direct3D = Microsoft.DirectX.Direct3D;
    
namespace Dodger
{
    class DodgerGame : Form
    {

        #region Секция переменных-членов класса DodgerGame
    
        // Ссылка на устройство
        Device device;
    
        // Ссылки на элементы дороги
        private Mesh roadMesh = null;
        private Material[] roadMaterials = null;
        private Texture[] roadTextures = null;
    
        // Объявления переменных-характеристик для управления дорогой
        public const float ROAD_LOCATION_LEFT = 2.5F;
        public const float ROAD_LOCATION_RIGHT = -2.5F;
        private const float ROAD_SIZE = 100.0F;
        private const float MAXIMUM_ROAD_SPEED = 250.0F;
        private const float ROAD_SPEED_INCREMENT = 0.5F;
    
        // Глубина рисуемой дороги
        private float RoadDepth0 = 0.0F;
        private float RoadDepth1 = -100.0F;
        private float RoadSpeed = 30.0F;
    
        // Точное время
        private float elapsedTime = Utility.Timer(DirectXTimer.GetElapsedTime);
    
        // Ссылка на объект автомобиля
        private Car car = null;
    
        // Ссылка на препятствия
        private Obstacles obstacles;
        // Высота препятствий
        private const float OBSTACLE_HEIGHT = Car.HEIGHT * 0.85f;
    
        // Переменные игры
        private bool isGameOver = true; // Стоп-игра 
        private int gameOverTick = 0; // Для обеспечения задержки
        private bool hasGameStarted = false; // Игра идет
        private int score = 0; // Очки 
    
        // Ссылки на шрифты 
        private Direct3D.Font scoreFont = null;
        private Direct3D.Font gameFont = null;
    
        // Информация о трех лучших достижениях
        private HighScore[] highScores = new HighScore[3];
        private string defaultHighScoreName = string.Empty;
    
        #endregion

        public DodgerGame()
        {
            this.Size = new Size(800, 600); // Размер окна формы
            this.Text = "Lab37. Игра Dodger"; // Заголовок окна
            this.SetStyle(ControlStyles.AllPaintingInWmPaint
                | ControlStyles.Opaque, true);
    
            // Чтение лучших результатов из системного реестра
            this.LoadHighScores();
        }

        // Функция создания и инициализации объектов DirectX
        public void InitializeGraphics()
        {
            // Установить параметры представления
            PresentParameters presentParams = new PresentParameters();
            presentParams.Windowed = true;
            presentParams.SwapEffect = SwapEffect.Discard;
    
            // Включить буфер глубины устройства
            presentParams.AutoDepthStencilFormat = DepthFormat.D16;
            presentParams.EnableAutoDepthStencil = true;
    
            // Установить адаптер программную поддержку по умолчанию
            int adapterOrdinal = Manager.Adapters.Default.Adapter;
            CreateFlags flags = CreateFlags.SoftwareVertexProcessing;
    
            // Проверка возможности использования прямой аппаратной поддержки
            Caps caps = Manager.GetDeviceCaps(adapterOrdinal, DeviceType.Hardware);
    
            // Возможна ли прямая аппаратная поддержка вершинного буфера
            if (caps.DeviceCaps.SupportsHardwareTransformAndLight)
                flags = CreateFlags.HardwareVertexProcessing;
    
            // Возможна ли прямая аппаратная поддержка устройства
            if (caps.DeviceCaps.SupportsPureDevice)
                flags |= CreateFlags.PureDevice;
    
            // Создать устройство
            device = new Device(adapterOrdinal, DeviceType.Hardware, 
                this, flags, presentParams);
    
            // Подписаться на событие регистрации сброса устройства
            device.DeviceReset += new EventHandler(OnDeviceReset);
    
            // Прямой вызов обработчика
            this.OnDeviceReset(device, null);
    
            // Сброс переменных игры при новой игре
            LoadDefaultGameOptions();
    
            // Регистрация обработчика щелчка мыши для остановки игры 
            this.Click += new EventHandler(DodgerGame_Click);
    
            // Создание шрифтов
            scoreFont = new Direct3D.Font(device,
                new System.Drawing.Font("Arial", 12.0f, 
                FontStyle.Bold));
    
            gameFont = new Direct3D.Font(device,
                new System.Drawing.Font("Arial", 36.0f, 
                FontStyle.Bold | FontStyle.Italic));
        }
    
        // Приостановка движения
        static bool stopGame = true;
        void DodgerGame_Click(object sender, EventArgs e)
        {
            if (stopGame)
            {
                Utility.Timer(DirectXTimer.Stop);
                stopGame = false;
            }
            else
            {
                Utility.Timer(DirectXTimer.Start);
                stopGame = true;
            }
        }
    
        // Обработчик события сброса устройства
        private void OnDeviceReset(object sender, EventArgs e)
        {
            // Установка камеры
            device.Transform.Projection = Matrix.PerspectiveFovLH(
                (float)Math.PI / 4, // Угол зрения равен 45 градусов
                (float)this.Width / (float)this.Height, // Форматное соотношение сторон
                1.0F, // Ближний план
                1000.0F); // Дальний план
            device.Transform.View = Matrix.LookAtLH(
                new Vector3(0.0F, 9.5F, 17.0F), // Положение камеры
                new Vector3(), // Положение объекта текущее
                new Vector3(0, 1, 0)); // Направление камеры
    
            // Установка фильтрации текстуры
            // Установка фильтра сжатия - анизотропный, если поддерживается
            // видеокартой, иначе - линейный (если тоже поддерживается)
            if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyAnisotropic)
                device.SamplerState[0].MinFilter = TextureFilter.Anisotropic;
            else if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyLinear)
                device.SamplerState[0].MinFilter = TextureFilter.Linear;
    
            // Установка фильтра растяжения - анизотропный, если поддерживается
            // видеокартой, иначе - линейный (если тоже поддерживается)
            if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyAnisotropic)
                device.SamplerState[0].MagFilter = TextureFilter.Anisotropic;
            else if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyLinear)
                device.SamplerState[0].MagFilter = TextureFilter.Linear;
    
            // Установка освещения
            // Поддерживаются ли несколько полноценных источников освещения
            if ((device.DeviceCaps.VertexProcessingCaps.SupportsDirectionAllLights) &&
                (device.DeviceCaps.MaxActiveLights > 1))
            {
                // Устанавливаем первый источник
                device.Lights[0].Type = LightType.Directional;
                device.Lights[0].Diffuse = Color.White;
                device.Lights[0].Direction = new Vector3(1, -1, -1);
                device.Lights[0].Commit();
                device.Lights[0].Enabled = true;
    
                // Устанавливаем второй источник
                device.Lights[1].Type = LightType.Directional;
                device.Lights[1].Diffuse = Color.White;
                device.Lights[1].Direction = new Vector3(-1, 1, -1);
                device.Lights[1].Commit();
                device.Lights[1].Enabled = true;
            }
            else
            {
                // Источники не поддерживаются, устанавливаем общий свет
                device.RenderState.Ambient = Color.White;
                // Настраиваем общий свет
                device.Lights[0].Type = LightType.Directional;
                device.Lights[0].Diffuse = Color.White;
                device.Lights[0].Direction = new Vector3(1, -1, -1);
                device.Lights[0].Commit();
                device.Lights[0].Enabled = true;
            }
    
            // Включить освещение сцены
            device.RenderState.Lighting = true;
    
            // Загрузка Mesh-файла и создание объекта дороги
            roadMesh = LoadMesh(device, @".\road.x", 
                ref roadMaterials, ref roadTextures);
    
            // Загрузка Mesh-файла и создание объекта автомобиля
            car = new Car(device);
    
            // Создание объекта-коллекции препятствий
            obstacles = new Obstacles();
        }
    
        // Код организации рендеринга
        protected override void OnPaint(PaintEventArgs e)
        {
            // Очистка экрана
            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer,
                Color.CornflowerBlue, 1.0F, 0);
    
            // Вычисление новых параметров рисования дороги,
            // автомобиля и препятствий 
            OnFrameUpdate();
    
            // Формирование о отображение сцены
            device.BeginScene();
            {
                // Отображение двух секций дороги с разной глубиной
                DrawRoad(0.0F, 0.0F, RoadDepth0);
                DrawRoad(0.0F, 0.0F, RoadDepth1);
    
                // Отображение автомобиля
                car.DrawCar(device);
    
                // Отображение препятствий
                foreach (Obstacle o in obstacles)
                {
                    o.DrawObstacle(device);
                }
    
                // Визуализация табло
                if (hasGameStarted)
                {
                    // Отображение очков
                    scoreFont.DrawText(string.Format("Текущие очки: {0}", score),
                        new Rectangle(5, 5, 0, 0), 
                        DrawTextFormat.NoClip, Color.Yellow);
                }
                if (isGameOver)
                {
                    // Если игра закончена, уведомить игрока
                    if (hasGameStarted)
                    {
                        gameFont.DrawText("Вы проиграли. Ха-ха-ха!",
                            new Rectangle(25, 45, 0, 0), 
                            DrawTextFormat.NoClip, Color.Red);
                    }
                    if ((System.Environment.TickCount - gameOverTick) >= 1000)
                    {
                        // Вывести подсказку - как начать игру, если после 
                        // остановки предыдущей игры прошло более 1 секунды
                        gameFont.DrawText("Нажмите клавишу Space",
                            new Rectangle(25, 130, 0, 0), 
                            DrawTextFormat.NoClip, Color.WhiteSmoke);
                    }
    
                    // Показ лучших достижений
                    gameFont.DrawText("Лучшие результаты: ", 
                        new Rectangle(25, 205, 0, 0),
                        DrawTextFormat.NoClip, Color.Red);
    
                    for (int i = 0; i < highScores.Length; i++)
                    {
                        gameFont.DrawText(string.Format("Игрок: {0} : {1}", 
                            highScores[i].Name,
                            highScores[i].Score),
                            new Rectangle(25, 270 + (i * 55), 0, 0), 
                            DrawTextFormat.NoClip, Color.Yellow);
                    }
                }
            }
            device.EndScene();
    
            device.Present();
            this.Invalidate();
        }

        // Функция загрузки Mesh-файла дороги
        public static Mesh LoadMesh(Device device, string file,
            ref Material[] meshMaterials, ref Texture[] meshTextures)
        {
            ExtendedMaterial[] mtrl;
    
            // Загружаем Mesh-файл
            Mesh tempMesh = Mesh.FromFile(file, MeshFlags.Managed, device, out mtrl);
    
            // Если имеется материал, сохраняем его
            if (mtrl != null && mtrl.Length > 0)
            {
                // Создаем массивы
                meshMaterials = new Material[mtrl.Length];
                meshTextures = new Texture[mtrl.Length];
    
                // Сохраняем каждый материал и текстуру
                for (int i = 0; i < mtrl.Length; i++)
                {
                    meshMaterials[i] = mtrl[i].Material3D;
                    if (mtrl[i].TextureFilename != null
                        && mtrl[i].TextureFilename != string.Empty)
                    {
                        // Пытаемся загрузить файл с текстурой
                        meshTextures[i] = TextureLoader.FromFile(device,
                            @".\" // Текущий каталог размещения сборки
                            + mtrl[i].TextureFilename);
                    }
                }
            }
    
            return tempMesh;
        }

        // Функция отображения секции дороги
        private void DrawRoad(float x, float y, float z)
        {
            device.Transform.World = Matrix.Translation(x, y, z);
                
            for (int i = 0; i < roadMaterials.Length; i++)
            {
                device.Material = roadMaterials[i];
                device.SetTexture(0, roadTextures[i]);
                roadMesh.DrawSubset(i);
            }
        }
    
        // Функция обновления кадра, привязанная
        // к таймеру повышенной точности
        private void OnFrameUpdate()
        {
            // Если игра остановлена, не выполнять функцию
            if (isGameOver || !hasGameStarted)
                return;
    
            // Извлекаем точное время
            elapsedTime = Utility.Timer(DirectXTimer.GetElapsedTime);
    
            // Вычисляем текущую глубину дороги
            RoadDepth0 += RoadSpeed * elapsedTime;
            RoadDepth1 += RoadSpeed * elapsedTime;
    
            // Проверяем необходимость смены секций дороги
            if (RoadDepth0 > 75.0F)
            {
                RoadDepth0 = RoadDepth1 - 100.0F;
                AddObstacles(RoadDepth0);
            }
            if (RoadDepth1 > 75.0F)
            {
                RoadDepth1 = RoadDepth0 - 100.0F;
                AddObstacles(RoadDepth1);
            }
    
            // Преверяем, что препятствие миновало автомобиль
            // Увеличиваем очки и усложняем игру
            Obstacles removeObstacles = new Obstacles();
            foreach (Obstacle o in obstacles)
            {
                if (o.Depth > car.Diameter - (Car.DEPTH / 2))
                {
                    removeObstacles.Add(o);
                    // Увеличить скорость дороги
                    RoadSpeed += ROAD_SPEED_INCREMENT;
    
                    // Проверить, чтобы скорость не зашкалила
                    if (RoadSpeed >= MAXIMUM_ROAD_SPEED)
                    {
                        RoadSpeed = MAXIMUM_ROAD_SPEED;
                    }
    
                    // Увеличить скорость горизонтальных перемещений автомобиля
                    car.IncrementSpeed();
    
                    // Увеличить очки за количество миновавших препятствий
                    score += (int)(RoadSpeed * (RoadSpeed / car.Speed));
                }
            }
    
            // Удалить миновавшие препятствия с помощью временного списка
            foreach (Obstacle o in removeObstacles)
            {
                obstacles.Remove(o);
                o.Dispose();// Вызвать сборщик мусора (упаковать)
            }
            removeObstacles.Clear();// Очистить временный список
    
            // Проверка столкновения и остановка игры
            foreach (Obstacle o in obstacles)
            {
                if (o.IsHittingCar(car.Location, car.Diameter))
                {
                    // Если обнаружено столкновение - закончить игру!
                    isGameOver = true; // Стоп-игра, перекур!
                    gameOverTick = System.Environment.TickCount;
                    // Останавливаем таймер
                    Utility.Timer(DirectXTimer.Stop);
    
                    // Проверить результат среди лучших игроков
                    CheckHighScore();
                }
            }
    
            // Вычисление координаты нового положения автомобиля
            car.LocationCar(elapsedTime);
    
            // Вычисление координаты нового положения препятствий
            foreach (Obstacle o in obstacles)
            {
                o.LocationObstacle(elapsedTime, RoadSpeed);
            }
        }
    
        // Переопределение виртуального метода OnKeyDown() в классе DodgerGame
        protected override void OnKeyDown(KeyEventArgs e)
        {
            // Выход по Esc - завершение работы приложения
            if (e.KeyCode == Keys.Escape)
                this.Close();
    
            // Блокировка клавиатуры
            if (System.Environment.TickCount - gameOverTick < 1000)
                return;
    
            // Перемещение влево
            if (e.KeyCode == Keys.Left || e.KeyCode == Keys.NumPad4)
            {
                car.IsMovingLeft = true;
                car.IsMovingRight = false;
            }
    
            // Перемещение вправо
            if (e.KeyCode == Keys.Right || e.KeyCode == Keys.NumPad6)
            {
                car.IsMovingRight = true;
                car.IsMovingLeft = false;
            }
    
            // Остановить перемещение автомобиля
            if (e.KeyCode == Keys.Down || e.KeyCode == Keys.NumPad2)
            {
                car.IsMovingLeft = false;
                car.IsMovingRight = false;
            }
    
            // Space - перезапуск игры
            if (e.KeyCode == Keys.Space && isGameOver)
            {
                LoadDefaultGameOptions();
                isGameOver = false;
                hasGameStarted = true;
            }
        }
    
        // Добавление препятствий в коллекцию Obstacles
        private void AddObstacles(float minDepth)
        {
            // Вычислить возможное число препятствий
            int numberToAdd = (int)((ROAD_SIZE / car.Diameter - 1) / 2.0f);
            // Получить зазор между препятствиями, достаточный для автомобиля
            float minSize = ((ROAD_SIZE / numberToAdd) - car.Diameter) / 2.0f;
    
            // Генерировать препятствия
            for (int i = 0; i < numberToAdd; i++)
            {
                // Случайным образом сгенерировать размер в допустимом диапазоне
                float depth = minDepth - ((float)Utility.Rnd.NextDouble() * minSize);
                // Поправка глубины
                depth -= (i * (car.Diameter * 2));
    
                // Выбрать левую или правую стороны дороги
                float location = (Utility.Rnd.Next(50) > 25) 
                    ? ROAD_LOCATION_LEFT : ROAD_LOCATION_RIGHT;
    
                // Добавить препятствие в коллекцию
                obstacles.Add(new Obstacle(device, location, OBSTACLE_HEIGHT, depth));
            }
        }
    
        // Установка переменных игры в начальные значения
        private void LoadDefaultGameOptions()
        {
            // Характеристики дороги
            RoadDepth0 = 0.0f;
            RoadDepth1 = -100.0f;
            RoadSpeed = 30.0f;
    
            // Характеристики автомобиля
            car.Location = ROAD_LOCATION_LEFT;
            car.Speed = 10.0f;
            car.IsMovingLeft = false;
            car.IsMovingRight = false;
    
            // Сбрасываем очки 
            score = 0;
    
            // Удаляем препятствия предыдущей игры
            foreach (Obstacle o in obstacles)
            {
                // Сначала освобождаем каждый ресурс 
                // по отдельности, затем удаляем
                o.Dispose();
            }
            obstacles.Clear();
    
            // Заполняем новыми препятствиями
            AddObstacles(RoadDepth1);
    
            // Стартуем наш таймер
            Utility.Timer(DirectXTimer.Start);
        }

        // Функции поддержки лучших результатов в истории игры
        // Если текущий игрок побеждает, то запоминает его
        // индекс и просит ввести свое имя
        private void CheckHighScore()
        {
            int index = -1;// Сбрасываем текущий индекс
            for (int i = highScores.Length - 1; i >= 0; i--)
            {
                if (score >= highScores[i].Score) // Нашли, кого заменить
                {
                    index = i;
                }
            }
    
            // Делаем замену, если индекс отличен от -1
            if (index >= 0)
            {
                // Освобождаем найденную позицию 
                for (int i = highScores.Length - 1; i > index; i--)
                {
                    highScores[i] = highScores[i - 1];
                }
                // Сохраняем на освободившемся месте текущий результат
                highScores[index].Score = score;
                highScores[index].Name = Input.InputBox("Вы получили высокий счет!!!",
                    "Просим ввести ваше имя:", defaultHighScoreName);
            }
        }
    
        // Сохранение лучших результатов в системном реестре
        public void SaveHighScores()
        {
            Microsoft.Win32.RegistryKey key =
                Microsoft.Win32.Registry.CurrentUser.CreateSubKey(
                "Software\\MDXBoox\\Dodger");
    
            try // Делаем попытку
            {
                for (int i = 0; i < highScores.Length; i++)
                {
                    key.SetValue(string.Format("Player{0}", i), highScores[i].Name);
                    key.SetValue(string.Format("Score{0}", i), highScores[i].Score);
                }
                key.SetValue("PlayerName", defaultHighScoreName);
            }
			finally // Выполняем при любом исходе
            {
                if (key != null)
                {
                    key.Close(); // Обязательно закрываем ключ
                }
            }
        }
    
        // Загрузка списка лучших результатов из системного реестра
        private void LoadHighScores()
        {
            Microsoft.Win32.RegistryKey key =
                Microsoft.Win32.Registry.CurrentUser.CreateSubKey(
                "Software\\MDXBoox\\Dodger");
    
            try // Делаем попытку
            {
                for (int i = 0; i < highScores.Length; i++)
                {
                    highScores[i].Name = (string)key.GetValue(
                        string.Format("Player{0}", i), string.Empty);
    
                    highScores[i].Score = (int)key.GetValue(
                        string.Format("Score{0}", i), 0);
                }
                defaultHighScoreName = (string)key.GetValue(
                    "PlayerName", System.Environment.UserName);
            }
			finally // Выполняем при любом исходе
            {
                if (key != null)
                {
                    key.Close(); // Обязательно закрываем ключ
                }
            }
        }
    }
    
    // Структура для сохранения лучших достижений игры
    public struct HighScore
    {
        private int realScore; // Результат
        private string playerName; // Игрок
    
        // Свойства для доступа к закрытым членам
        public int Score
        {
            get
            {
                return realScore;
            }

            set
            {
                realScore = value;
            }
        }
    
        public string Name
        {
            get
            {
                return playerName;
            }
            set
            {
                playerName = value;
            }
        }
    }
}
